"""Declares the :class:`~DTO` and :class:`~ImmutableDTO` classes."""
import copy


class DTO(dict):
    """A dict-like collection of properties."""
    __immutable__ = False

    @classmethod
    def fromdict(cls, mapping, with_dto=False):
        """Create a new Data Transfer Object (DTO) from a regular
        Python dictionary.
        """
        return cls(mapping).as_dto(with_dto=with_dto)

    def as_dict(self):
        """Convert the DTO  to a dictionary. For use in environments
        where the DTO can not be serialized and no hooks are available
        to implement it.
        """
        dto = copy.deepcopy(self)
        for key in dto.keys():
            value = dto[key]
            if not isinstance(value, (DTO, ImmutableDTO)):
                continue
            dto[key] = value.as_dict()
        return dict(dto)

    def as_dto(self, with_dto=False):
        """Convert all dictionaries in the DTO to DTO
        instances.
        """
        dto = copy.deepcopy(self)
        dto_cls = type(self)
        for k in dto.keys():
            value = dto[k]
            if with_dto and hasattr(value, 'as_dto'):
                dto[k] = value.as_dto(with_dto)
                continue

            if isinstance(value, dict):
                dto[k] = dto_cls(**value).as_dto(with_dto=with_dto)
                continue

            if isinstance(value, (list, set, tuple)):
                cls = type(value)
                dto[k] = cls([dto_cls.fromdict(x) if isinstance(x, dict) else x
                    for x in value])

        return dto

    def __serialize__(self, func):
        return func(dict(self))

    def __getattr__(self, attname):
        if attname not in self:
            raise AttributeError(attname)
        return self[attname]

    def __setattr__(self, attname, value):
        if self.__immutable__:
            raise TypeError(attname)
        self[attname] = value

    def __repr__(self):
        super_repr = dict.__repr__(self)
        return "<%s: %s>" % (type(self).__name__, super_repr)

    def __sub__(self, keys):
        if not isinstance(keys, list):
            raise TypeError("Can only create new DTO with list.")
        return type(self)({x: self[x] for x in self.keys() if x not in keys})

    def __or__(self, dto):
        if not isinstance(dto, dict):
            raise ValueError("Cannot union dict with non-dict type.")
        new = type(self)(self)
        new.update(dto)
        return new

    def __and__(self, dto):
        """Return the intersection of `self` and `dto` by their keys."""
        keys = self.__getkeys(dto)
        if isinstance(dto, dict):
            params = {x: dto[x] for x in keys if x in self}
        else:
            params = {x: self[x] for x in keys if x in self}
        return type(self)(params)

    @staticmethod
    def __getkeys(dto):
        return set(dto) if not isinstance(dto, dict)\
            else set(dto.keys())


class ImmutableDTO(DTO):
    """Like :class:`DTO`, but the properties are immutable."""
    __immutable__ = True



# pylint: skip-file
import typing

from .patternset import PatternSet


__all__ = ['matches']


def matches(
    patterns: typing.Union[str, typing.Set[str] ],
    values: typing.Union[str, typing.Set[str] ],
    sep: str = '.'
) -> set:
    """Matches `patterns` against `values` using the separator character
    `sep`. The input parameters may be either strings or iterables containing
    strings. Return a set containing the matches.
    """
    if isinstance(patterns, str):
        patterns = {patterns}
    patterns = PatternSet(patterns, sep)
    if isinstance(values, str):
        values = {values}
    return patterns & values



"""Declares base exception classes that all Unimatrix projects should
recognize, and may raise to set an error condition with a well-defined
meaning.
"""
import functools
import uuid


class FailedOperation(Exception):
    """Represents an exception that was caused by a failure
    of a requested operation.
    """
    version = "unimatrixone.io/v1"
    default_status = 421
    default_code = 'UNKNOWN_ERROR'

    @classmethod
    def catch(cls, on_exception):
        """Returns a decorator that invokes callable `on_exception`
        when a :class:`FailedOperation` is raised.
        """
        def decorator(func):
            @functools.wraps(func)
            def f(*args, **kwargs): #pylint: disable=invalid-name
                try:
                    return func(*args, **kwargs)
                except FailedOperation as e: #pylint: disable=invalid-name
                    return on_exception(e)
            return f
        return decorator

    def __init__(self, status=None, code=None, stderr=None):
        super().__init__()
        self.code = code or self.default_code
        self.status = status or self.default_status
        self.stderr = stderr or ''
        self.uid = str(uuid.uuid4())

    def get_metadata(self):
        """Returns metadata describing the exception instance."""
        return {
            'uid': self.uid,
            'annotations': {
                'unimatrixone.io/http-status-code': self.status
            }
        }

    @property
    def dto(self):
        """Returns the exception as a Data Transfer Object (DTO) that is
        ready for serialization and transmission over the network.
        """
        return {
            'api_version': self.version,
            'kind': type(self).__name__,
            'metadata': self.get_metadata(),
            'spec': {
                'code': self.code
            }
        }


class AccessDenied(FailedOperation):
    """Raised when an operation fails due to request not
    having permission.
    """
    default_code = 'ACCESS_DENIED'
    default_status = 401


class ResourceDoesNotExist(FailedOperation):
    """Raised when an operation fails due to an upstream
    resource not existing.
    """
    default_code = 'DOES_NOT_EXIST'
    default_status = 404


class Unreachable(FailedOperation):
    """Raised when an operation fails due to an upstream
    resource not being reachable.
    """
    default_code = 'UNREACHABLE'
    default_status = 503


class Timeout(Unreachable):
    """Raised when an operation fails due to a timeout
    when accessing an upstream resource.
    """



#pylint: skip-file
import types
import unittest

from ..test import TestCaseSelector


class BaseTestSelectionTestCase:

    def get_class(self, skip, *args, **kwargs):
        scopes = kwargs.pop('scopes', None)
        selector = TestCaseSelector(*args, **kwargs)

        @selector.scopes(scopes)
        class mock_class:
            def test_mock_method(*args):
                if skip:
                    self.fail("This method should be skipped.")
        return mock_class

    def get_method(self, skip, *args, **kwargs):
        scopes = kwargs.pop('scopes', None)
        selector = TestCaseSelector(*args, **kwargs)

        class mock_class:
            @selector.scopes(scopes)
            def test_mock_method(*args):
                if skip:
                    self.fail("This method should be skipped.")
        return mock_class().test_mock_method

    def assertTestSkipped(self, testcase):
        if isinstance(testcase, type):
            return self.assertTrue(getattr(testcase, '__unittest_skip__', True))
        try:
            testcase()
        except unittest.SkipTest:
            pass
        else:
            self.fail("Test was ran but should be skipped.")

    def assertTestNotSkipped(self, testcase):
        if isinstance(testcase, type):
            return self.assertFalse(getattr(testcase, '__unittest_skip__', False))
        try:
            testcase()
        except unittest.SkipTest:
            self.fail("Test was skipped but should be ran.")

    def assertTestDecorator(self, skip, factory, *args, **kwargs):
        decoratable = factory(skip, *args, **kwargs)
        return self.assertTestSkipped(decoratable)\
            if skip else self.assertTestNotSkipped(decoratable)



"""Annotations are used to provide additional metadata about resources,
class instances and other datastructures.
"""
#pylint: disable=bad-whitespace


DJANGO_DOMAIN               = 'django.python.unimatrixone.io'
DJANGO_PROJECT_NAME         = '%s/project' % DJANGO_DOMAIN
DJANGO_SETTINGS_MODULE      = '%s/settings' % DJANGO_DOMAIN
PLATFORM_DOMAIN             = "platform.unimatrixone.io"
PLATFORM_LANGUAGE_NAME      = "%s/language-name" % PLATFORM_DOMAIN
PLATFORM_LANGUAGE_VERSION   = "%s/language-version" % PLATFORM_DOMAIN
PLATFORM_FRAMEWORK_NAME     = "%s/framework-name" % PLATFORM_DOMAIN
PLATFORM_FRAMEWORK_VERSION  = "%s/framework-version" % PLATFORM_DOMAIN
PLATFORM_OS_NAME            = "%s/os-name" % PLATFORM_DOMAIN
PLATFORM_OS_RELEASE         = "%s/os-release" % PLATFORM_DOMAIN



#pylint: skip-file
import unittest

from ..test import TestCaseSelector
from .basetestselection import BaseTestSelectionTestCase


class TestScopeSelectionTestCase(unittest.TestCase, BaseTestSelectionTestCase):

    def test_class_is_not_skipped_when_current_scope_in_scopes(self):
        self.assertTestDecorator(False, self.get_class, 'foo', scopes=['bar'],
            current_stage='foo', current_scope='bar')

    def test_class_is_not_skipped_when_current_scope_in_multiple_scopes(self):
        self.assertTestDecorator(False, self.get_class, 'foo',
            scopes=['bar', 'baz'], current_stage='foo', current_scope='bar')

    def test_class_is_skipped_when_scopes_are_not_defined(self):
        self.assertTestDecorator(True, self.get_class, 'foo', scopes=None,
            current_stage='foo', current_scope='bar')

    def test_class_is_skipped_when_current_scope_is_not_in_defined_scopes_empty(self):
        self.assertTestDecorator(True, self.get_class, 'foo', scopes=[],
            current_stage='foo', current_scope='bar')

    def test_class_is_skipped_when_current_scope_is_not_in_defined_scopes(self):
        self.assertTestDecorator(True, self.get_class, 'foo', scopes=['baz'],
            current_stage='foo', current_scope='bar')

    def test_class_is_skipped_when_current_scope_is_none(self):
        self.assertTestDecorator(True, self.get_class, 'foo', scopes=['baz'],
            current_stage='foo', current_scope=None)

    def test_scope_decorator_skips(self):
        selector = TestCaseSelector('foo', current_stage='foo', scopes=[],
            current_scope='bar')

        @selector.scope('bar')
        def f(*args, **kwargs):
            pass

        self.assertTestNotSkipped(f)



# pylint: skip-file
import unittest

from ..test import AsyncTestCase


class AsyncTestCaseMethodTestCase(AsyncTestCase):

    @unittest.expectedFailure
    async def test_can_run_async_test_method(self):
        self.fail()



class AsyncTestCaseSetupTestCase(AsyncTestCase):

    async def setUp(self):
        self.foo = 1

    def test_setup_is_ran_async(self):
        self.assertTrue(hasattr(self, 'foo'))

    async def tearDown(self):
        self.assertTrue(self.foo == 1)
